/* tarot implements the rules of the tarot game
 * Copyright (C) 2019  Vivien Kraus
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef H_TAROT_FEATURES_INCLUDED
#define H_TAROT_FEATURES_INCLUDED

#include <stddef.h>
#include <tarot/bid.h>
#include <tarot/cards.h>

#ifdef __cplusplus
extern "C"
{
#endif

  struct TarotGame;

  struct TarotGameEvent;

  /* All features at once */

  /**
   * tarot_features_arff_header:
   * @encoded: (type boolean):
   * @returns: (transfer none):
   */
  const char *tarot_features_arff_header (int encoded);

  /**
   * tarot_features_arff_data:
   * @base: (type TarotGame):
   * @added: (type TarotGameEvent):
   * @encoded: (type boolean):
   * @dest_out: (array length=max):
   */
  size_t tarot_features_arff_data (const struct TarotGame *base, const struct TarotGameEvent *added, int score, int encoded, size_t max, char *dest_out);

  /**
   * tarot_features_arff_data_alloc:
   * @base: (type TarotGame):
   * @added: (type TarotGameEvent):
   * @encoded: (type boolean):
   * @returns: (transfer full):
   */
  char *tarot_features_arff_data_alloc (const struct TarotGame *base, const struct TarotGameEvent *added, int score, int encoded);

  /* General-purpose features */

  /**
   * tarot_features_variant:
   * @game: (type TarotGame):
   * @n_players: (out):
   * @with_call: (type boolean) (out):
   */
  void tarot_features_variant (const struct TarotGame *game, size_t *n_players, int *with_call);

  /**
   * tarot_features_position:
   * @game: (type TarotGame):
   * @i_trick: (out): the number of tricks played so far
   * @position: (out): the number of players that play before me (for
   * bids, decls, or for the current trick)
   */
  void tarot_features_position (const struct TarotGame *game, size_t *i_trick, size_t *position);

  /**
   * tarot_features_partner_known:
   * @game: (type TarotGame):
   * @known: (out) (type boolean):
   */
  void tarot_features_partner_known (const struct TarotGame *game, int *known);

  /* Features specific to bidding */

  /**
   * tarot_features_bid_minimum:
   * @game: (type TarotGame):
   * @minimum: (out):
   */
  void tarot_features_bid_minimum (const struct TarotGame *game, TarotBid *minimum);

  /**
   * tarot_features_count_faces:
   * @game: (type TarotGame):
   * @n_kings: (out):
   * @n_marriages: (out):
   * @n_extended_marriages: (out):
   * @n_queens: (out):
   * @n_knights: (out):
   */
  void tarot_features_count_faces (const struct TarotGame *game,
				   size_t *n_kings,
				   size_t *n_marriages,
				   size_t *n_extended_marriages,
				   size_t *n_queens,
				   size_t *n_knights);

  /**
   * tarot_features_count_minor:
   * @game: (type TarotGame):
   * @n_shortest: (out):
   * @n_second_shortest: (out):
   * @n_second_longest: (out):
   * @n_longest: (out):
   */
  void tarot_features_count_minor (const struct TarotGame *game,
				   size_t *n_shortest,
				   size_t *n_second_shortest,
				   size_t *n_second_longest,
				   size_t *n_longest);
  
  /**
   * tarot_features_count_oudlers:
   * @game: (type TarotGame):
   * @has_petit: (out) (type boolean):
   * @has_excuse: (out) (type boolean):
   * @has_twentyone: (out) (type boolean):
   */
  void tarot_features_count_oudlers (const struct TarotGame *game,
				     int *has_petit,
				     int *has_excuse,
				     int *has_twentyone);

  /**
   * tarot_features_count_trumps:
   * @game: (type TarotGame):
   * @n_trumps: (out):
   * @max_trump: (out):
   * @min_trump: (out):
   * @q1_trumps: (out):
   * @q3_trumps: (out):
   * @median_trumps: (out):
   */
  void tarot_features_count_trumps (const struct TarotGame *game,
				    size_t *n_trumps,
				    TarotNumber *max_trump,
				    TarotNumber *min_trump,
				    TarotNumber *q1_trumps,
				    TarotNumber *q3_trumps,
				    TarotNumber *median_trumps);

  /**
   * tarot_strategy_bid:
   * @game: (type TarotGame):
   * @event: (type TarotGameEvent):
   * @bid: (out):
   */
  void tarot_strategy_bid (const struct TarotGame *game,
			   const struct TarotGameEvent *event,
			   TarotBid *bid);

  /**
   * tarot_strategy_outbid:
   * @game: (type TarotGame):
   * @event: (type TarotGameEvent):
   * @outbid: (out):
   */
  void tarot_strategy_outbid (const struct TarotGame *game,
			      const struct TarotGameEvent *event,
			      int *outbid);

  /**
   * tarot_strategy_call_by_length:
   * @game: (type TarotGame):
   * @event: (type TarotGameEvent):
   * @shortest: (out) (type boolean):
   * @second_shortest: (out) (type boolean):
   * @second_longest: (out) (type boolean):
   * @longest: (out) (type boolean):
   */
  void tarot_strategy_call_by_length (const struct TarotGame *game,
				      const struct TarotGameEvent *event,
				      int *shortest,
				      int *second_shortest,
				      int *second_longest,
				      int *longest);

  /**
   * tarot_strategy_call_myself:
   * @game: (type TarotGame):
   * @event: (type TarotGameEvent):
   * @myself: (out) (type boolean):
   */
  void tarot_strategy_call_myself (const struct TarotGame *game,
				   const struct TarotGameEvent *event,
				   int *myself);

  /**
   * tarot_strategy_call_marriage:
   * @game: (type TarotGame):
   * @event: (type TarotGameEvent):
   * @will_own_marriage: (out) (type boolean):
   */
  void tarot_strategy_call_marriage (const struct TarotGame *game,
				     const struct TarotGameEvent *event,
				     int *will_own_marriage);

  /**
   * tarot_strategy_discard_save:
   * @game: (type TarotGame):
   * @event: (type TarotGameEvent):
   * @n_queens_saved: (out): number of queens for which we don't plan to own the king
   * @n_knights_saved: (out):
   * @n_jacks_saved: (out):
   */
  void tarot_strategy_discard_save (const struct TarotGame *game,
				    const struct TarotGameEvent *event,
				    size_t *n_queens_saved,
				    size_t *n_knights_saved,
				    size_t *n_jacks_saved);

  /**
   * tarot_strategy_discard_junk:
   * @game: (type TarotGame):
   * @event: (type TarotGameEvent):
   * @n_lesser_cards_left_behind: (out):
   */
  void tarot_strategy_discard_junk (const struct TarotGame *game,
				    const struct TarotGameEvent *event,
				    size_t *n_lesser_cards_left_behind);

  /**
   * tarot_strategy_discard_by_length:
   * @game: (type TarotGame):
   * @event: (type TarotGameEvent):
   * @from_shortest: (out): number of cards discarded from the shortest suit with the dog
   * @from_second_shortest: (out):
   * @from_second_longest: (out):
   * @from_longest: (out):
   */
  void tarot_strategy_discard_by_length (const struct TarotGame *game,
					 const struct TarotGameEvent *event,
					 size_t *from_shortest,
					 size_t *from_second_shortest,
					 size_t *from_second_longest,
					 size_t *from_longest);

  /**
   * tarot_features_team:
   * @game: (type TarotGame):
   * @attack: (out) (type boolean): whether we may be an attacker
   * @defence: (out) (type boolean): whether we may be a defendent
   */
  void tarot_features_team (const struct TarotGame *game,
			    int *attack,
			    int *defence);

  /**
   * tarot_features_trick_players:
   * @game: (type TarotGame):
   * @n_already_played: (out):
   * @n_allies_after_last_opponent: (out):
   */
  void tarot_features_trick_players (const struct TarotGame *game,
				     size_t *n_already_played,
				     size_t *n_allies_after_last_opponent);

  /**
   * tarot_features_minor_suit:
   * @game: (type TarotGame):
   * @ally_may_be_master: (out) (type boolean): whether noone can play
   * a stronger card of the minor suit than the current taker and the
   * current taker is an ally, or whether an ally who has still not
   * played yet may play the master card of the minor suit.  Same for
   * trumps, but we don't consider the possibility that someone
   * trumps.  That's why we have counted how many cards of the suit 
   * are remaining.
   * @ennemy_may_be_master: (out) (type boolean):
   * @called_suit: (out) (type boolean):
   * @n_times_lead: (out):
   * @n_remaining_cards: (out):
   * @n_remaining_points: (out):
   */
  void tarot_features_minor_suit (const struct TarotGame *game,
				  int *called_suit,
				  size_t *n_times_lead,
				  size_t *n_remaining_cards,
				  size_t *n_remaining_points,
				  int *ally_may_be_master,
				  int *ennemy_may_be_master);

  /**
   * tarot_features_trump:
   * @game: (type TarotGame):
   * @n_remaining: (out):
   * @ally_may_be_master: (out) (type boolean):
   * @ennemy_may_be_master: (out) (type boolean):
   */
  void tarot_features_trump (const struct TarotGame *game,
			     size_t *n_remaining,
			     int *ally_may_be_master,
			     int *ennemy_may_be_master);

  /**
   * tarot_features_points:
   * @game: (type TarotGame):
   * @n_halfpoints: (out):
   * @n_oudlers: (out):
   */
  void tarot_features_points (const struct TarotGame *game,
			      size_t *n_halfpoints,
			      size_t *n_oudlers);

  /**
   * tarot_strategy_lead:
   * @game: (type TarotGame):
   * @card: (type TarotGameEvent):
   * @called_suit: (out) (type boolean):
   * @suit_length: (out):
   * @n_times_lead: (out):
   * @trumps: (out) (type boolean):
   */
  void tarot_strategy_lead (const struct TarotGame *game,
			    const struct TarotGameEvent *card,
			    int *called_suit,
			    size_t *suit_length,
			    size_t *n_times_lead,
			    int *trumps);

  /**
   * tarot_strategy_purity:
   * @game: (type TarotGame):
   * @event: (type TarotGameEvent):
   * @n_smaller_kept: (out):
   * @n_greater_nonface_kept: (out):
   * @n_points_kept: (out):
   * @n_greater_kept: (out):
   */
  void tarot_strategy_purity (const struct TarotGame *game,
			      const struct TarotGameEvent *event,
			      size_t *n_smaller_kept,
			      size_t *n_greater_nonface_kept,
			      size_t *n_points_kept,
			      size_t *n_greater_kept);

  /**
   * tarot_strategy_excuse:
   * @game: (type TarotGame):
   * @card: (type TarotGameEvent):
   * @excuse: (out) (type boolean):
   */
  void tarot_strategy_excuse (const struct TarotGame *game,
			      const struct TarotGameEvent *card,
			      int *excuse);

  /**
   * tarot_strategy_trump_master:
   * @game: (type TarotGame):
   * @card: (type TarotGameEvent):
   * @actually_master: (out) (type boolean):
   * @obvious_master: (out) (type boolean):
   */
  void tarot_strategy_trump_master (const struct TarotGame *game,
				    const struct TarotGameEvent *card,
				    int *actually_master,
				    int *obvious_master);

  /**
   * tarot_strategy_play_points:
   * @game: (type TarotGame):
   * @card: (type TarotGameEvent):
   * @n_halfpoints: (out):
   * @oudler: (out) (type bool):
   */
  void tarot_strategy_play_points (const struct TarotGame *game,
				   const struct TarotGameEvent *card,
				   size_t *n_halfpoints,
				   int *oudler);

  /**
   * tarot_features_strongest_player:
   * @with_call: (type boolean):
   * @owners: (array length=n_cards):
   * @confidence: (out):
   */
  TarotPlayer tarot_features_strongest_player (size_t n_players,
					       int with_call,
					       size_t n_cards,
					       const TarotPlayer *owners,
					       double *confidence);

  struct TarotPredictor;
  typedef struct TarotPredictor TarotPredictor;
  
  struct TarotGameEvent;

  struct TarotGame;

  /**
   * tarot_predictor_construct:
   * @mem_out: (array length=max_mem):
   * @alignment: (out):
   */
  size_t tarot_predictor_construct (size_t max_mem, char *mem_out, size_t *alignment);

  void tarot_predictor_copy (TarotPredictor *predictor, const TarotPredictor *source);

  /**
   * tarot_predictor_alloc: (constructor)
   */
  TarotPredictor *tarot_predictor_alloc (void);

  TarotPredictor *tarot_predictor_dup (const TarotPredictor *predictor);

  void tarot_predictor_free (TarotPredictor *predictor);

  /**
   * tarot_predictor_set_game:
   * @game: (type TarotGame):
   */
  void tarot_predictor_set_game (TarotPredictor *predictor, const struct TarotGame *game);


  /**
   * tarot_predictor_best:
   * @score: (out):
   * @returns: (nullable) (type Tarot.GameEvent) (transfer none):
   */
  const struct TarotGameEvent *tarot_predictor_best (const TarotPredictor *predictor, double *score);

  /**
   * tarot_predictor_eval:
   * @score_out: (array length=max):
   * @candidates: (array length=n_candidates) (type TarotGameEvent*) (transfer none):
   */
  size_t tarot_predictor_eval (const TarotPredictor *predictor, size_t n_candidates, struct TarotGameEvent **candidates,
			       size_t start, size_t max, double *score_out);

  /**
   * tarot_features_cnn_add:
   * @data_out: (array length=max):
   */
  size_t tarot_features_cnn_add (const TarotGame *game, size_t max, double *data_out);

  /**
   * tarot_features_cnn_alloc:
   * @data: (out) (transfer full) (array length=n):
   * @n: (out):
   */
  void tarot_features_cnn_alloc (const TarotGame *game, double **data, size_t *n);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* not H_TAROT_FEATURES_INCLUDED */
