/* tarot implements the rules of the tarot game
 * Copyright (C) 2019  Vivien Kraus
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "program_commons.h"

static char *default_seed = "default";

static size_t n_players = 4;
static int with_call = 0;
static char **seed = &default_seed;

static void parse_options (int argc, char *argv[]);

int
main (int argc, char *argv[])
{
  TarotGame *game = NULL;
  TarotGameEvent *setup_event;
  TarotGameEvent *deal_event;
  char *output;
  tarot_program_setup ();
  parse_options (argc, argv);
  game = tarot_game_alloc ();
  setup_event = tarot_game_event_alloc_setup (n_players, with_call);
  deal_event = tarot_game_event_alloc_deal_all_random (n_players, strlen (*seed), *seed);
  if (tarot_game_add_event (game, setup_event) != TAROT_GAME_OK)
    {
      fprintf (stderr, _ ("Error: could not set the game up for %lu players (call: %d)\n"),
	       n_players, with_call);
      return EXIT_FAILURE;
    }
  if (tarot_game_add_event (game, deal_event) != TAROT_GAME_OK)
    {
      fprintf (stderr, _ ("Error: could not use this random seed, '%s', to deal.  Please retry with a different seed.\n"),
	       *seed);
      return EXIT_FAILURE;
    }
  output = tarot_game_save_to_xml_alloc (game);
  printf ("%s", output);
  return EXIT_SUCCESS;
}

static void
parse_options (int argc, char *argv[])
{
  int c;
  char *end;
  while (1)
    {
      static struct option long_options[] =
        {
	 {N_ ("help"), no_argument, 0, 'h'},
	 {N_ ("version"), no_argument, 0, 'v'},
	 {N_ ("n-players"), required_argument, 0, 'n'},
	 {N_ ("call-allowed"), no_argument, 0, 'c'},
	 {N_ ("no-call-allowed"), no_argument, &with_call, 0},
	 {N_ ("seed"), required_argument, 0, 's'},
	 {0, 0, 0, 0}
        };
      int option_index = 0;
      for (option_index = 0; long_options[option_index].name != NULL; option_index++)
	{
	  long_options[option_index].name = _ (long_options[option_index].name);
	}
      option_index = 0;
      c = getopt_long (argc, argv, "hvn:cs:", long_options, &option_index);
      if (c == -1)
        break;
      switch (c)
        {
        case 0:
          break;

	case 'h':
	  printf (_ ("Usage: tarot-deal [OPTIONS]...\n\
Create a new game with a random deal for all players.  Options:\n\
- -h, --help: print this message and exit\n\
- -v, --version: print the program version and exit\n\
- -n N_PLAYERS, --n-players=N_PLAYERS: set the number of players\n\
   (default: %lu)\n\
- -c, --call-allowed: the game allows a call\n\
-     --no-call-allowed: the game does not allow a call (default)\n\
- -s, --seed=\"SEED\": set the seed for the random deal.  SEED must\n\
   not be empty.\n\
\n\
If the seed leads to a petit sec, then the deal fails.  So in case of\n\
a failure you may want to retry for a few different seeds.\n\
"), n_players);
	  exit (EXIT_SUCCESS);
	  break;

	case 'v':
	  printf (_ ("%s (libtarot %s)\n\
\n\
Copyright © 2019 Vivien Kraus\nThis is free software; see the source for copying conditions.  There is NO warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"),
		  "tarot-deal",
		  tarot_libtarot_version ());
	  exit (EXIT_SUCCESS);
	  break;

        case 'n':
	  n_players = strtoul (optarg, &end, 10);
	  if (strcmp (end, optarg) == 0)
	    {
	      fprintf (stderr, _ ("Error: the number of players should be a number, not '%s'.\n"),
		       optarg);
	      exit (EXIT_FAILURE);
	    }
	  else if (strcmp (end, "") != 0)
	    {
	      fprintf (stderr, _ ("Error: the number of players (%lu) has an unexpected suffix, '%s': '%s'.\n"),
		       n_players, end, optarg);
	      exit (EXIT_FAILURE);
	    }
          break;

        case 'c':
	  with_call = 1;
          break;

        case 's':
	  *seed = strdup (optarg);
	  if (strlen (*seed) == 0)
	    {
	      fprintf (stderr, _ ("Error: the seed must not be empty.  The default is usually \"default\".\n"));
	      exit (EXIT_FAILURE);
	    }
          break;

        case '?':
          /* getopt_long already printed an error message. */
	  exit (EXIT_FAILURE);
          break;

        default:
          abort ();
        }
    }
  if (optind < argc)
    {
      fprintf (stderr, _ ("Error: extra arguments:\n"));
      while (optind < argc)
        fprintf (stderr, _ ("- %s\n"), argv[optind++]);
      exit (EXIT_FAILURE);
    }
}
